<%

Const PageSize = 4096
Const HeapStepCount = 4

Const TopStatLimit = 5
Const FragThreshold = 75

Dim I
Dim TopHeapCount
Dim TopStatCount
Dim TotalReserve
Dim TotalCommit
Dim HighFragHeaps
Dim Message
Dim Recommendation
Dim SummaryCount

Function AnalyzeAndReportHeapInfo()
'''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Primary function called to analyze each heap in the '
' dump file currently loaded						  '
'''''''''''''''''''''''''''''''''''''''''''''''''''''''
	SummaryCount = Manager.Results.Count

	Manager.Write "<h1><a name='HeapReport" & g_UniqueReference & "'>Heap Analysis</a></h1><br>"
	
	g_Progress.CurrentStatus = "Loading heap information. Please wait..."
	g_Progress.SetCurrentRange 0, HeapStepCount
	
	g_HeapArray = g_HeapInfo.SortHeapsByReserved()
	For I = 0 to g_HeapInfo.Count - 1
		TotalReserve = TotalReserve + g_HeapArray(I).ReservedMemory
		TotalCommit = TotalCommit + g_HeapArray(I).CommittedMemory
	Next
	
	%>
	
	<h2>Heap Summary</h2>
	<table class=myCustomText ID="Table3">
		<tr>
			<td><b>Number of heaps</b></td>
			<td><b>&nbsp;&nbsp;<%=g_HeapInfo.Count%> Heaps</b></td>
		</tr>
		<tr>
			<td><b>Total reserved memory</b></td>
			<td><b>&nbsp;&nbsp;<%=PrintMemory(TotalReserve)%></b></td>
		</tr>
		<tr>
			<td><b>Total committed memory</b></td>
			<td><b>&nbsp;&nbsp;<%=PrintMemory(TotalCommit)%></b></td>
		</tr>
	</table>
	<br><br>
	<%
	If g_HeapInfo.Count > TopHeapLimit Then
		TopHeapCount = TopHeapLimit
	Else
		TopHeapCount = g_HeapInfo.Count
	End If
	
	g_Progress.CurrentPosition = 1
	
	Manager.Write "<h2>Top " & TopHeapCount & " heaps by reserved memory</h2><br>"
	
	Set g_HeapGraph = new BarGraph
	g_HeapGraph.SetRowCount TopHeapCount
	
	For I = 0 to TopHeapCount - 1
		g_HeapGraph.Rows(I).Value = g_HeapArray(I).ReservedMemory
		g_HeapGraph.Rows(I).Caption = g_Debugger.GetAs32BitHexString(g_HeapArray(I).Handle)
		g_HeapGraph.Rows(I).Caption2 = PrintMemory(g_HeapArray(I).ReservedMemory)
		g_HeapGraph.Rows(I).Link = GetHeapLink(g_HeapArray(I))
	Next
	g_HeapGraph.DrawGraph
	
	g_Progress.CurrentPosition = 2
	
	g_Progress.CurrentStatus = "Sorting heaps by committed memory"
	g_HeapArray = g_HeapInfo.SortHeapsByCommitted()
	
	Manager.Write "<h2>Top " & TopHeapCount & " heaps by committed memory</h2><br>"
	
	Set g_HeapGraph = new BarGraph
	g_HeapGraph.SetRowCount TopHeapCount
	
	For I = 0 to TopHeapCount - 1
		g_HeapGraph.Rows(I).Value = g_HeapArray(I).CommittedMemory
		g_HeapGraph.Rows(I).Caption = g_Debugger.GetAs32BitHexString(g_HeapArray(I).Handle)
		g_HeapGraph.Rows(I).Caption2 = PrintMemory(g_HeapArray(I).CommittedMemory)
		g_HeapGraph.Rows(I).Link = GetHeapLink(g_HeapArray(I))
	Next
	g_HeapGraph.DrawGraph
	
	g_Progress.CurrentPosition = 3
	
	g_Progress.CurrentStatus = "Generating detailed heap report"
	g_Progress.SetCurrentRange 0, g_HeapInfo.Count
	
	Manager.Write "<h2>Heap Details</h2><br><br>"
	
	HighFragHeaps = ""
	For each g_Heap in g_HeapInfo
		PrintHeapInfo g_Heap
		g_Progress.CurrentPosition = g_Heap.Index + 1
	Next
	
	If HighFragHeaps <> "" Then
		Message = "Detected symptoms of high fragmentation in the following heaps in " & g_ShortDumpFileName &_
		"<br><br>" & HighFragHeaps
		
		Manager.ReportWarning Message, "Heap fragmentation is often caused by one of the following two reasons<br><br>" &_
		"1. Small heap memory blocks that are leaked (allocated but never freed) over time<br>" &_
		"2. Mixing long lived small allocations with short lived long allocations<br><br>" &_
		"Both of these reasons can prevent the NT heap manager from using free memory effeciently since they are " &_
		" spread as small fragments that cannot be used as a single large allocation" 
	End If
		
	g_Progress.CurrentPosition = 4
	
End Function 'AnalyzeAndReportHeapInfo

Sub ShowHeapInfoNoneDetectedIfNecessary
    ' Check this after checking leaktrack data, so we don't report "no heap issues" along with some 
    ' leaktrack warnings in the summary
	If SummaryCount = Manager.Results.Count Then
		Manager.ReportInformation "<br>DebugDiag did not detect any known <b>native heap(unmanaged)</b> problems in " & g_ShortDumpFileName & " using the current set of scripts.<br><br>"
	End If
End Sub ' ShowHeapInfoNoneDetectedIfNecessary

Sub PrintHeapInfo(ByVal Heap)
'''''''''''''''''''''''''''''''''''''''''''''''''
' Prints out detail info for the heap passed in '
'''''''''''''''''''''''''''''''''''''''''''''''''
	Dim Segment
	Dim UnCommitted, Reserved, Committed, LargestFreeBlock
	Dim PctFragmentation
	Dim BusyStats, J
	Dim HeapName
	Dim SegmentError, SegmentRecommendation
	Dim HeapOwner, HeapModule
	
	UnCommitted = Heap.ReservedMemory - Heap.CommittedMemory
	
	If UnCommitted > 0 Then
		PctFragmentation = ((UnCommitted -  Heap.LargestUncommittedRange)/UnCommitted) * 100
	Else
		PctFragmentation = 0
	End If
	
	If PctFragmentation > FragThreshold Then
		HighFragHeaps = HighFragHeaps & "<br><a href='" & GetHeapLink(Heap) & "'><b>" &_
						g_Debugger.GetAs32BitHexString(Heap.Handle) &  "</b></a> ("
		
		If Heap.Name <> "" Then
			HighFragHeaps = HighFragHeaps & "<b>" & Heap.Name & "</b> - "
		End If
								
		HighFragHeaps = HighFragHeaps & FormatNumber(PctFragmentation , 2) & "% Fragmented)"
	End If
	
	HeapName = Heap.Name
	%>
	<h4>
		<a name='#<%= g_UniqueReference & Heap.Handle%>'>
			Heap <%=Heap.Index%> - <%=g_Debugger.GetAs32BitHexString(Heap.Handle)%>
		</a>
	</h4>
	<table class=myCustomText ID="Table1">
		<% If HeapName <> "" Then %>
		<tr>
			<td><b>Heap Name</b></td>
			<td><b>&nbsp;&nbsp;<%=HeapName%></b></td>
		</tr>
		<tr>
			<td><b>Heap Description</b></td>
			<td><b>&nbsp;&nbsp;<%=Heap.Description%></b></td>
		</tr>
		<% End If %>
		<tr>
			<td><b>Reserved memory</b></td>
			<td><b>&nbsp;&nbsp;<%=PrintMemory(Heap.ReservedMemory)%></b></td>
		</tr>
		<% If Heap.ReservedMemory > 0 Then %>
		<tr>
			<td><b>Committed memory</b></td>
			<td><b>&nbsp;&nbsp;<%=PrintMemory(Heap.CommittedMemory)%>
				(<%= FormatNumber((Heap.CommittedMemory / Heap.ReservedMemory) * 100, 2)%>% of reserved)
			</b></td>
		</tr>
		<tr>
			<td><b>Uncommitted memory</b></td>
			<td><b>&nbsp;&nbsp;<%=PrintMemory(UnCommitted)%>
				(<%= FormatNumber((UnCommitted / Heap.ReservedMemory) * 100, 2)%>% of reserved)
			</b></td>
		</tr>
		<% End If %>
		<tr>
			<td><b>Number of heap segments</b></td>
			<td><b>&nbsp;&nbsp;<%=Heap.Count%> segments</b></td>
		</tr>
		<tr>
			<td>Number of uncommitted ranges</td>
			<td>&nbsp;&nbsp;<%=Heap.NumberOfUnCommittedRanges%> range(s)</td>
		</tr>
		<tr>
			<td>Size of largest uncommitted range</td>
			<td>&nbsp;&nbsp;<%=PrintMemory(Heap.LargestUncommittedRange)%></td>
		</tr>
		<tr>
			<td>Calculated heap fragmentation</td>
			<td>&nbsp;&nbsp;<%=FormatNumber(PctFragmentation , 2)%>%</td>
		</tr>
	</table>
	<br><br>
	<h5>Segment Information</h5>
	<table class=myCustomText cellspacing=3 ID="Table2">
		<tr>
			<th>Base Address</th>
			<th>Reserved Size</th>
			<th>Committed Size</th>
			<th>Uncommitted Size</th>
			<th>Number of uncommitted ranges</th>
			<th>Largest uncommitted block</th>
			<th>Calculated heap fragmentation</th>
		</tr>
	<%
	If Heap.ReservedMemory > 0 Then
	If Heap.Count > 63 AND FormatNumber((Heap.CommittedMemory / Heap.ReservedMemory) * 100, 2) < 10 AND Not UCase(HeapName) = "DEFAULT PROCESS HEAP" Then
		SegmentError = "In <b>" & g_ShortDumpFileName & "</b>, heap <a href='" & GetHeapLink(Heap) & "'>" & g_Debugger.GetAs32BitHexString(Heap.Handle) & "</a> contains <b>" & Heap.Count & "</b> segments, with very little committed memory in the heap. " &_
						"Any allocation request for this heap greater than <b>" & PrintMemory(Heap.LargestUncommittedRange) & "</b> will fail." &_
						"<br><br>Memory statistics for this heap:<br><br>" &_
						"Reserved Memory: <b>" & PrintMemory(Heap.ReservedMemory) & "</b><br>" &_
						"Committed Memory: <b>" & PrintMemory(Heap.CommittedMemory) & "</b> (" & FormatNumber((Heap.CommittedMemory / Heap.ReservedMemory) * 100, 2) & "% of reserved)"
		If HeapName <> "" Then
			SegmentError = SegmentError & "<br>Heap Name: <b>" & HeapName & "</b>"
			HeapOwner = GetHeapOwnerModule(HeapName)
		End If
		
		SegmentRecommendation = "Check with the heap creator "
		
		If HeapName = Empty OR HeapOwner = "<Unable to obtain heap owner data>" Then
			SegmentRecommendation = SegmentRecommendation & "(unable to obtain heap owner data for this heap), "
		Else
			Set HeapModule = g_ModuleCache.GetModuleByName(HeapOwner)
			If Not HeapModule Is Nothing Then
				SegmentRecommendation = SegmentRecommendation & "(<b>" & HeapModule.VSCompanyName & "</b>), "
			Else
				SegmentRecommendation = SegmentRecommendation & "(created by: <b>" & HeapOwner & "</b>), "
			End If
		End If
		
		SegmentRecommendation = SegmentRecommendation & "and ensure the heap is destroyed when all outstanding allocations have been freed. " &_
									"<br><br>If there are outstanding allocations, review the allocations to see if a leak is preventing this heap from being destroyed."
		
		Manager.ReportWarning SegmentError, SegmentRecommendation
	End If
	End If
	
	For each Segment in Heap
		Reserved = Segment.NumberOfPages * PageSize
		Uncommitted = Segment.NumberOfUnCommittedPages * PageSize
		Committed = Reserved - Uncommitted
		LargestFreeBlock = Segment.LargestUnCommittedRange
		
		If UnCommitted > 0 Then
			PctFragmentation = ((UnCommitted - LargestFreeBlock)/UnCommitted) * 100
		Else
			PctFragmentation = 0
		End If
		
		%>
		<tr>
			<td><%=g_Debugger.GetAs32BitHexString(Segment.BaseAddress)%></td>
			<td><%=PrintMemory(Reserved)%></td>
			<td><%=PrintMemory(Committed)%></td>
			<td><%=PrintMemory(Uncommitted)%></td>
			<td><%=Segment.NumberOfUnCommittedRanges%></td>
			<td><%=PrintMemory(LargestFreeBlock)%></td>
			<td><%=FormatNumber(PctFragmentation, 2)%>%</td>
		</tr>
		<%
	Next
	%>
	</table>
	<br><br>
	<%
	BusyStats = Heap.BusyStatisticsBySize
	
	If UBound(BusyStats) > TopStatLimit Then
		TopStatCount = TopStatLimit
	Else
		TopStatCount = UBound(BusyStats) + 1
	End If
	
	If TopStatCount > 0 Then
		Manager.Write "<h5>Top " & TopStatCount & " allocations by size</h5><br>"		
		Set g_HeapGraph = new BarGraph
		g_HeapGraph.SetRowCount TopStatCount
		
		For J = 0 to TopStatCount - 1
			'Manager.Write "<b>" & J + 1 & "</b> Allocation Size - " & BusyStats(J,0)
			'Manager.Write PrintMemory(BusyStats(J,0) * BusyStats(J,1))
			g_HeapGraph.Rows(J).Value = BusyStats(J,0) * BusyStats(J,1)
			g_HeapGraph.Rows(J).Caption = "Allocation Size - " & BusyStats(J,0)
			g_HeapGraph.Rows(J).Caption2 = PrintMemory(BusyStats(J,0) * BusyStats(J,1))
		Next
		g_HeapGraph.DrawGraph
	End If
	
	BusyStats = Heap.BusyStatisticsByCount
	
	If UBound(BusyStats) > TopStatLimit Then
		TopStatCount = TopStatLimit
	Else
		TopStatCount = UBound(BusyStats) + 1
	End If
	
	If TopStatCount > 0 Then
		Manager.Write "<h5>Top " & TopStatCount & " allocations by count</h5><br>"		
		Set g_HeapGraph = new BarGraph
		g_HeapGraph.SetRowCount TopStatCount
		
		For J = 0 to TopStatCount - 1
			'Manager.Write "<b>" & J + 1 & "</b> Allocation Size - " & BusyStats(J,0)
			'Manager.Write BusyStats(J,1) & " allocation(s)"
			g_HeapGraph.Rows(J).Value = BusyStats(J,1)
			g_HeapGraph.Rows(J).Caption = "Allocation Size - " & BusyStats(J,0)
			g_HeapGraph.Rows(J).Caption2 = BusyStats(J,1) & " allocation(s)"
		Next
		g_HeapGraph.DrawGraph
	End If
	
	Manager.Write "<a href='#'>Back to Top</a>"
	%>
	<br><br><br><br>
	<%
End Sub 'PrintHeapInfo

Function GetHeapLink(ByVal Heap)
	GetHeapLink = "#" & g_UniqueReference & Heap.Handle
End Function 'GetHeapLink

Function IsHeapFunction(ByVal Address)
'''''''''''''''''''''''''''''''''''''''''''''
' Determines if the function at the address '
' passed in is a heap funtion or not		'
'''''''''''''''''''''''''''''''''''''''''''''

	Dim FunctionName
	
	FunctionName = Split(UCase(GetFunctionName(Address)), "+", -1)
	
	Select Case FunctionName(0)
		Case "NTDLL!RTLFREEHEAP", "NTDLL!RTLALLOCATEHEAP", "NTDLL!RTLPCOALESCEFREEBLOCKS"
			IsHeapFunction = True
		Case "NTDLL!RTLPINTERLOCKEDPOPENTRYSLIST", "NTDLL!RTLPALLOCATEFROMHEAPLOOKASIDE", "NTDLL!EXPINTERLOCKEDPOPENTRYSLISTFAULT"
			IsHeapFunction = True
		Case "NTDLL!RTLPDPHREPORTCORRUPTEDBLOCK", "NTDLL!RTLPBREAKPOINTHEAP", "NTDLL!RTLPDPHISNORMALHEAPBLOCK"
			IsHeapFunction = True
		Case "NTDLL!RTLPEXTENDHEAP", "NTDLL!RTLPFINDANDCOMMITPAGES"
			IsHeapFunction = True
		Case Else
			IsHeapFunction = False
	End Select
End Function 'IsHeapFunction

Function AnalyzeHeapCorruption(ByVal ExceptionThread)
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
' Call this in cases where the AV was caused by heap corruption AND pageheap was enabled. This will walk the stack '
' and return the module info from the frame that called the initial memory-related function.					   '
' *** Be sure to add any known heap functions that we want to ignore to 'KnownHeapFuncs' ***							   '
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

	Dim StackFrames, Module
	Dim I, StackFrame
	Dim KnownHeapFuncs, StringFunc
	
	KnownHeapFuncs = "NTDLL!RTLFREEHEAP;NTDLL!RTLALLOCATEHEAP;NTDLL!RTLPCOALESCEFREEBLOCKS;NTDLL!RTLPINTERLOCKEDPOPENTRYSLIST;" &_
					"NTDLL!RTLPINTERLOCKEDPOPENTRYSLIST;NTDLL!RTLPALLOCATEFROMHEAPLOOKASIDE;NTDLL!RTLPDPHREPORTCORRUPTEDBLOCK;NTDLL!RTLPBREAKPOINTHEAP;" &_
					"NTDLL!RTLPDPHISNORMALHEAPBLOCK;NTDLL!RTLPDPHNORMALHEAPFREE;NTDLL!RTLPDEBUGPAGEHEAPFREE;NTDLL!RTLDEBUGFREEHEAP;" &_
					"NTDLL!RTLFREEHEAPSLOWLY;MSVCRT!FREE"
					
	Set StackFrames = ExceptionThread.StackFrames
	For I = 0 to StackFrames.Count - 1
		Set StackFrame = StackFrames(I)
		StringFunc = Split(UCase(GetSymbolFromAddress(StackFrame.InstructionAddress)), "+", -1)
		If InStr(1, KnownHeapFuncs, UCase(StringFunc(0))) = 0 Then
			Set AnalyzeHeapCorruption = GetModuleFromAddress(StackFrame.InstructionAddress)
			Exit Function
		End If
	Next
End Function 'AnalyzeHeapCorruption

Function GetHeapOwnerModule(ByVal HeapName)
	Dim Name

	If InStr(1, UCase(HeapName), "MICROSOFT DATA ACCESS RUNTIME MPHEAP") > 0 Then
		GetHeapOwnerModule = "msado15"
	ElseIf InStr(1, UCase(HeapName), "ODBC32 MPHEAP") > 0 Then
		GetHeapOwnerModule = "odbc32"
	ElseIf InStr(1, UCase(HeapName), "MICROSOFT XML PARSER V3 MPHEAP") > 0 Then
		GetHeapOwnerModule = "MSXML3"
	ElseIf InStr(1, UCase(HeapName), "!") > 0 Then
		Name = Split(UCase(HeapName), "!", -1)
		GetHeapOwnerModule = Name(0)
	Else
		GetHeapOwnerModule = "<Unable to obtain heap owner data>"
	End If
End Function 'GetHeapOwnerModule

Function GetGlobalFlagDescription(ByVal GFValue)
	Dim colGFlags, key
	Set colGFlags = CreateObject("Scripting.Dictionary")
	
	colGFlags.Add &H00000001, "Stop on exception"
	colGFlags.Add &H00000002, "Show loader snaps"
	colGFlags.Add &H00000004, "Debug initial command"
	colGFlags.Add &H00000008, "Stop on a hung GUI"
	colGFlags.Add &H00000010, "Enable heap tail checking"
	colGFlags.Add &H00000020, "Enable heap free checking"
	colGFlags.Add &H00000040, "Enable heap parameter checking"
	colGFlags.Add &H00000080, "Enable heap validation on call"
	colGFlags.Add &H00000100, "Enable pool tail checking"
	colGFlags.Add &H00000200, "Enable pool free checking"
	colGFlags.Add &H00000400, "Enable pool tagging"
	colGFlags.Add &H00000800, "Enable heap tagging"
	colGFlags.Add &H00001000, "Create user-mode stack trace DB"
	colGFlags.Add &H00002000, "Create kernel-mode stack trace DB"
	colGFlags.Add &H00004000, "Maintain a list of objects for each type"
	colGFlags.Add      32768, "Enable heap tagging by DLL" ' &H8000
	colGFlags.Add &H00020000, "Enable debugging of Microsoft Win32 subsystem"
	colGFlags.Add &H00040000, "Enable loading of kernel debugger symbols"
	colGFlags.Add &H00080000, "Disable paging of kernel stacks"
	colGFlags.Add &H00100000, "Enable critical system breaks"
	colGFlags.Add &H00200000, "Disable heap coalesce on free"
	colGFlags.Add &H00400000, "Enable close exception"
	colGFlags.Add &H00800000, "Enable exception logging"
	colGFlags.Add &H01000000, "Enable object handle type tagging"
	colGFlags.Add &H02000000, "Place heap allocations at ends of pages"
	colGFlags.Add &H04000000, "Debug WINLOGON"
	colGFlags.Add &H08000000, "Disable kernel-mode DbgPrint and KdPrint output"
	colGFlags.Add &H80000000, "Disable protected DLL verification"
	
	For Each key In colGFlags
		If (GFValue And key) = key Then
			GetGlobalFlagDescription = GetGlobalFlagDescription & ", " & colGFlags(key)
		End If
	Next
	
	' Drop the first comma
	GetGlobalFlagDescription = Right(GetGlobalFlagDescription, Len(GetGlobalFlagDescription) - 2)
	
	If GetGlobalFlagDescription = "" Then GetGlobalFlagDescription = "Unknown"
End Function 'GetGlobalFlagDescription

Function IsPageHeapEnabled(ByVal GFValue)
	' Not checking the other pageheap-related bits like &H10, &H20, &H40, &H80
	' Only checking 0x02000000 ("Place heap allocations at ends of pages"), which is what you get from 'full' pageheap
	IsPageHeapEnabled = ((GFValue And &H02000000) = &H02000000)
End Function
%>